# ruff: noqa: F401
from typing import Annotated

from anthropic.types import (
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
)
from openai.types.chat import ChatCompletionChunk
from pydantic import Field

from .chunk import ChunkEvent
from .output import OutputEvent
from .start import StartEvent

# Create a type alias for Anthropic events
AnthropicEvent = (
    RawContentBlockStartEvent |
    RawContentBlockDeltaEvent |
    RawContentBlockStopEvent |
    RawMessageStartEvent |
    RawMessageDeltaEvent |
    RawMessageStopEvent
)

# Create a type alias for OpenAI events (in the future we may have more than one)
OpenAIEvent = ChatCompletionChunk

# Create a discriminated union of all possible event types.
# Pydantic will use the 'type' field to determine which model to use.
Event = Annotated[
    StartEvent | OutputEvent | ChunkEvent,
    Field(discriminator="type"),
]



from pathlib import Path
from typing import Any, Literal

import structlog

from .base import BaseOperation

logger = structlog.get_logger("timbal.server.fs.operations.write")


class WriteOperation(BaseOperation):
    """Write content to file."""
    type: Literal["write"] = "write"
    content: str = ""

    async def __call__(self, base_path: Path) -> dict[str, Any]:
        """Execute the write operation."""
        try:
            full_path = base_path / self.path
            
            full_path.parent.mkdir(parents=True, exist_ok=True)
            full_path.write_text(self.content, encoding="utf-8")
            logger.info("file_written", path=self.path)
            return {"success": True, "message": f"File written: {self.path}"}
            
        except Exception as e:
            logger.error("write_file_failed", path=self.path, error=str(e))
            return {"error": f"Failed to write file: {str(e)}"}


import asyncio
import copy
from collections.abc import AsyncGenerator, Generator
from typing import Any

import structlog
from anthropic.types import (
    InputJSONDelta,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    TextDelta,
    ToolUseBlock,
)
from pydantic import BaseModel, ConfigDict
from uuid_extensions import uuid7

from ..types.events import AnthropicEvent, OpenAIEvent
from ..types.events.base import BaseEvent as TimbalBaseEvent
from ..types.events.chunk import ChunkEvent as TimbalChunkEvent
from ..types.events.output import OutputEvent as TimbalOutputEvent
from ..types.events.start import StartEvent as TimbalStartEvent

logger = structlog.get_logger("timbal.core.stream")


async def sync_to_async_gen(gen: Generator[Any, None, None], loop: asyncio.AbstractEventLoop) -> AsyncGenerator[Any, None]:
    """Auxiliary function to convert a sync generator to an async generator."""
    while True:
        # StopIteration is special in Python. It's used to implement generator protocol and can't
        # be pickled/transferred across threads properly. By catching it explicitly in the executor 
        # function and converting it to a sentinel value, we avoid problematic exception propagation.
        def _next():
            try:
                return next(gen)
            except StopIteration: 
                return None
        value = await loop.run_in_executor(None, _next)
        if value is None:
            break
        yield value


class AsyncGenState(BaseModel):
    """Tracks state for async generators during flow execution.

    Allow for arbitrary types not properly handled by Pydantic.
    Also allow for extra fields to be added to the model.
    """
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    gen: AsyncGenerator[Any, None] | None = None
    """The async generator being processed."""
    input: Any | None = None
    """The input for the step."""
    collections: Any = []
    """Accumulated results from generator yields (e.g. LLM message chunks)."""
    usage: dict[str, int] = {}
    """Resource usage stats (e.g. token counts for LLM calls)."""
    events_source: str | None = None # Allow None for late initializations.
    """The source of the events (e.g. 'openai', 'anthropic', 'timbal', 'str', ...)."""

    def collect(self) -> Any:
        return self.collections


def handle_openai_event(
    openai_event: OpenAIEvent,
    async_gen_state: AsyncGenState,
) -> Any:
    """Process OpenAI streaming events and update the async generator state.

    Handles different types of OpenAI events including:
    - Usage statistics (token counts)
    - Tool calls (function calling)
    - Text content deltas

    Args:
        openai_event: The OpenAI event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        Text chunk if the event contains content or tool call info, None otherwise
    """

    # TODO Review. Gemini sends usage stats in every chunk. It doesn't send a separate chunk with the total usage.
    # ? We guess they send the revised number in the last chunk.
    if openai_event.usage:
        openai_model = openai_event.model
        openai_usage = openai_event.usage

        input_tokens = int(openai_usage.prompt_tokens)
        input_tokens_details = openai_usage.prompt_tokens_details
        if hasattr(input_tokens_details, "cached_tokens"):
            input_cached_tokens = int(input_tokens_details.cached_tokens)
            if input_cached_tokens:
                input_tokens -= input_cached_tokens
                async_gen_state.usage[f"{openai_model}:input_cached_tokens"] = input_cached_tokens
        if hasattr(input_tokens_details, "audio_tokens"):
            input_audio_tokens = int(input_tokens_details.audio_tokens)
            if input_audio_tokens:
                input_tokens -= input_audio_tokens
                async_gen_state.usage[f"{openai_model}:input_audio_tokens"] = input_audio_tokens
        # ? We've seen this in some responses. Usually they return the image tokens as regular text tokens.
        # if hasattr(input_tokens_details, "image_tokens"):
        #     input_image_tokens = int(input_tokens_details.image_tokens)
        #     if input_image_tokens:
        #         input_tokens -= input_image_tokens
        #         async_gen_state.usage[f"{openai_model}:input_image_tokens"] = input_image_tokens
        # Text tokens are used as the default.
        # if hasattr(input_tokens_details, "text_tokens"):

        async_gen_state.usage[f"{openai_model}:input_text_tokens"] = input_tokens

        output_tokens = int(openai_usage.completion_tokens)
        output_tokens_details = openai_usage.completion_tokens_details
        if hasattr(output_tokens_details, "audio_tokens"):
            output_audio_tokens = int(output_tokens_details.audio_tokens)
            if output_audio_tokens:
                output_tokens -= output_audio_tokens
                async_gen_state.usage[f"{openai_model}:output_audio_tokens"] = output_audio_tokens

        async_gen_state.usage[f"{openai_model}:output_text_tokens"] = output_tokens

    if not len(openai_event.choices):
        return None

    if openai_event.choices[0].delta.tool_calls:
        tool_call = copy.deepcopy(openai_event.choices[0].delta.tool_calls[0])
        if tool_call.id:
            chunk_result = {
                "type": "tool_use",
                "id": tool_call.id,
                "name": tool_call.function.name,
                "input": ""
            }
            async_gen_state.collections.append(chunk_result)
        else:
            # ? Gemini doesn't add an id to the tool call.
            if not async_gen_state.collections:
                tool_use_id = f"call_{uuid7(as_type='str').replace('-', '')}"
                async_gen_state.collections.append({
                    "type": "tool_use",
                    "id": tool_use_id,
                    "name": tool_call.function.name,
                    "input": ""
                })
            async_gen_state.collections[-1]["input"] += tool_call.function.arguments
        return None

    if openai_event.choices[0].delta.content:
        text_chunk = openai_event.choices[0].delta.content
        chunk_result = {
            "type": "text",
            "text": text_chunk,
        }

        # Extra is allowed in the async gen pydantic model.
        if hasattr(openai_event, "citations"):
            async_gen_state.citations = openai_event.citations

        if async_gen_state.collections:
            async_gen_state.collections[-1]["text"] += text_chunk
        else:
            async_gen_state.collections.append(chunk_result)
        return text_chunk


def handle_anthropic_event(
    anthropic_event: AnthropicEvent,
    async_gen_state: AsyncGenState,
) -> Any:
    """Process Anthropic streaming events and update the async generator state.

    Handles different types of Anthropic events including:
    - Message start/usage info
    - Content block starts (text or tool use)
    - Content deltas (text or JSON)
    - Message completion with usage stats

    Args:
        anthropic_event: The Anthropic event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        Text chunk if the event contains content or tool input, None otherwise
    """

    # The raw content block stop event is sent when the LLM sends multiple blocks.
    # The raw message stop event is sent when the LLM is actually done.
    # if isinstance(anthropic_event, RawContentBlockStopEvent | RawMessageStopEvent):
    #     pass

    # TODO Handle other types of tokens.
    if isinstance(anthropic_event, RawMessageStartEvent):
        if anthropic_event.message.usage:
            # Output and delta messages do not include the model. 
            # We store this temp attribute to create the correct usage key underneath.
            anthropic_model = anthropic_event.message.model
            async_gen_state.anthropic_model = anthropic_model
            input_tokens_key = f"{anthropic_model}:input_tokens"
            input_tokens = anthropic_event.message.usage.input_tokens
            existing_input_tokens = async_gen_state.usage.get(input_tokens_key, 0)
            async_gen_state.usage[input_tokens_key] = existing_input_tokens + input_tokens
        return None

    if isinstance(anthropic_event, RawContentBlockStartEvent):
        if isinstance(anthropic_event.content_block, ToolUseBlock):
            chunk_result = {
                "type": "tool_use",
                "id": anthropic_event.content_block.id,
                "name": anthropic_event.content_block.name,
                "input": ""
            }
            async_gen_state.collections.append(chunk_result)
        else:
            chunk_result = {
                "type": "text",
                "text": anthropic_event.content_block.text
            }
            async_gen_state.collections.append(chunk_result)
        return None

    if isinstance(anthropic_event, RawContentBlockDeltaEvent):
        if isinstance(anthropic_event.delta, InputJSONDelta):
            text_chunk = anthropic_event.delta.partial_json
            async_gen_state.collections[-1]["input"] += text_chunk
            return None
        elif isinstance(anthropic_event.delta, TextDelta):
            text_chunk = anthropic_event.delta.text
            async_gen_state.collections[-1]["text"] += text_chunk
            return text_chunk

    if isinstance(anthropic_event, RawMessageDeltaEvent):
        if anthropic_event.usage:
            anthropic_model = async_gen_state.anthropic_model
            output_tokens_key = f"{anthropic_model}:output_tokens"
            output_tokens = anthropic_event.usage.output_tokens
            existing_output_tokens = async_gen_state.usage.get(output_tokens_key, 0)
            async_gen_state.usage[output_tokens_key] = existing_output_tokens + output_tokens
        return None


def handle_timbal_event(
    timbal_event: TimbalBaseEvent,
    async_gen_state: AsyncGenState,
) -> dict[str, Any] | None:
    """Process Timbal flow events and update the async generator state.

    Handles different types of Timbal events including:
    - Step start events (filtered out for parent flows)
    - Step chunk events (currently filtered)
    - Step output events (captures usage statistics)
    - Flow output events (captures final results)

    Args:
        timbal_event: The Timbal event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        None - Timbal events currently don't return streaming content
    """

    # For a parent flow, we don't need to stream the subflow step start events.
    if isinstance(timbal_event, TimbalStartEvent):
        return None

    # TODO Think how should we stream partial flow outputs results or chunks. 
    # Perhaps we should only stream up if the step output is selected as a flow output.
    if isinstance(timbal_event, TimbalChunkEvent):
        return None
    
    # TODO Think how should we stream partial flow outputs results or chunks. 
    # Perhaps we should only stream up if the step output is selected as a flow output.
    if isinstance(timbal_event, TimbalOutputEvent):
        if timbal_event.usage:
            for k, v in timbal_event.usage.items():
                current_key_value = async_gen_state.usage.get(k, 0)
                async_gen_state.usage[k] = current_key_value + v

        async_gen_state.collections = timbal_event.output
        return None

    return None


def handle_event(
    event: Any,
    async_gen_state: AsyncGenState,
) -> dict[str, Any] | None:
    """Route events to their appropriate handlers based on type.

    Dispatches events to specialized handlers for OpenAI, Anthropic, and Timbal events.
    Provides basic handling for string events and default fallback for other types.

    Args:
        event: The event to process
        async_gen_state: State object tracking the async generator execution

    Returns:
        Processed event content if available, None otherwise
    """
    if isinstance(event, OpenAIEvent):
        if async_gen_state.events_source is None:
            async_gen_state.events_source = "openai"
        if async_gen_state.events_source != "openai":
            raise ValueError("We cannot handle mixed events sources.")
        return handle_openai_event(event, async_gen_state)

    if isinstance(event, AnthropicEvent):
        if async_gen_state.events_source is None:
            async_gen_state.events_source = "anthropic"
        if async_gen_state.events_source != "anthropic":
            raise ValueError("We cannot handle mixed events sources.")
        return handle_anthropic_event(event, async_gen_state)

    if isinstance(event, TimbalBaseEvent):
        if async_gen_state.events_source is None:
            async_gen_state.events_source = "timbal"
        if async_gen_state.events_source != "timbal":
            raise ValueError("We cannot handle mixed events sources.")
        return handle_timbal_event(event, async_gen_state)

    if isinstance(event, str):
        if async_gen_state.events_source is None:
            async_gen_state.events_source = "str"
        if async_gen_state.events_source != "str":
            raise ValueError("We cannot handle mixed events sources.")
        if async_gen_state.collections:
            async_gen_state.collections += event
        else: 
            async_gen_state.collections = event
        return event

    # ? Implement custom behavior for other event types.

    # Default behavior for any other event type.
    async_gen_state.events_source = "any"
    async_gen_state.collections.append(event)
    return event



from typing import Any

from pydantic import BaseModel, ConfigDict

from ..models import dump
from ...state.context import RunContext


class BaseEvent(BaseModel):
    """Base class for all timbal events yielded during flow execution."""
    # Allow storing extra fields in the model and make immutable after creation.
    model_config = ConfigDict(extra="allow", frozen=True)

    type: str
    """The type of the event. This will be very useful for serializing and deserializing events."""
    run_id: str
    """The id of the run this event was emitted from."""
    path: str
    """The path of the element that yielded this event."""
    
    @classmethod
    async def build(cls, **data):
        """Factory method to create an event with pre-computed dump."""
        instance = cls(**data)
        # Compute and store the dump
        object.__setattr__(instance, '_dump', await dump(instance))
        return instance
    
    @property
    def dump(self) -> dict[str, Any]:
        """Pre-computed dump of the event."""
        if hasattr(self, '_dump'):
            return self._dump
        else:
            raise RuntimeError("Event was not created using BaseEvent.build() - use BaseEvent.build() instead of direct instantiation")



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, ClassVar

from pydantic import BaseModel, field_validator


class BaseOperation(BaseModel, ABC):
    """Base class for all filesystem operation messages."""
    type: str
    path: str
    
    # This will be set by the server when it starts
    # All operations automatically inherit the same base path
    # No need to pass context through every validation call
    _base_path: ClassVar[Path | None] = None

    @field_validator('path')
    @classmethod
    def validate_path(cls, v: str) -> str:
        """Validate that path is within allowed directory."""
        if cls._base_path is None:
            return v
            
        try:
            full_path = cls._base_path / v
            full_path.resolve().relative_to(cls._base_path.resolve())
            return v
        except ValueError as e:
            raise ValueError("Path outside allowed directory") from e

    @abstractmethod
    async def __call__(self, base_path: Path) -> dict[str, Any]:
        """Execute the operation with the given base path."""
        pass



from pathlib import Path
from typing import Any, Literal

import structlog

from .base import BaseOperation

logger = structlog.get_logger("timbal.server.fs.operations.mkdir")


class MkdirOperation(BaseOperation):
    """Create a directory."""
    type: Literal["mkdir"] = "mkdir"

    async def __call__(self, base_path: Path) -> dict[str, Any]:
        """Execute the mkdir operation."""
        try:
            full_path = base_path / self.path
            
            if full_path.exists():
                return {"error": f"Directory already exists: {self.path}"}
                
            full_path.mkdir(parents=True, exist_ok=True)
            logger.info("directory_created", path=self.path)
            return {"success": True, "message": f"Directory created: {self.path}"}
            
        except Exception as e:
            logger.error("create_directory_failed", path=self.path, error=str(e))
            return {"error": f"Failed to create directory: {str(e)}"}


from pydantic import BaseModel, ConfigDict

from ....types.field import Field
from ....types.message import Message


class BaseAgentParamsModel(BaseModel):
    """Base fixed parameter model for Agents. Used for internal validation."""
    model_config = ConfigDict(extra="ignore")

    system_prompt: str | None = Field(
        default=None,
        description="The system prompt to use for the agent."
    )
    model: str | None = Field(
        default="gpt-4.1-nano",
        description="The model to use for the agent."
    )
    max_tokens: int | None = Field(
        default=None,
        description="The maximum number of tokens to generate."
    )
    stream: bool | None = Field(
        default=False,
        description="Whether to stream the output."
    )


class AgentParamsModel(BaseAgentParamsModel):
    """Fixed parameter model for Agents."""

    prompt: Message = Field(description="The prompt to use for the agent.")


agent_params_model_schema = AgentParamsModel.model_json_schema()



"""
Google Drive Directory Listing

This module provides functionality to list and explore contents of Google Drive,
including both personal and shared drives. It offers detailed file information
and flexible search options.

Features:
---------
- List files and folders in any drive location
- Search in personal drive, shared drives, or both
- Get detailed file information:
  - File name and ID
  - File size (human-readable format)
  - Owner information
  - Last modified date
  - Sharing status
  - File type (folder/file)
- Save listings to a file
- Filter by scope (personal/shared/all)

Prerequisites:
-------------
1. Google Cloud Project setup:
   - Create a project at https://console.cloud.google.com
   - Enable the Google Drive API:
     https://console.cloud.google.com/flows/enableapi?apiid=drive.googleapis.com

2. Authentication:
   - Configure OAuth 2.0 credentials
   - Save credentials as 'credentials_gdrive.json'
   - For service accounts:
     https://cloud.google.com/iam/docs/service-accounts-create

Usage Examples:
--------------
1. List root directory:
   >>> list_directory()

2. List specific folder:
   >>> list_directory(folder="My Projects")

3. List shared drive contents:
   >>> list_directory(shared_drives="Team Drive")

4. List only shared files:
   >>> list_directory(scope="shared")

5. Save listing to file:
   >>> list_directory(output_path="drive_listing.json")

Note:
-----
- File sizes are displayed in human-readable format (B, KB, MB, GB)
- Dates are shown in local timezone
- Requires appropriate permissions for shared drives
- Maximum of 1000 items per listing
"""

from datetime import datetime
from pathlib import Path
import json
from typing import Any

from googleapiclient.discovery import build

from ...types import Field
from .authenticate import load_api_key, load_credentials


def format_size(
    size: int | None = Field(
        default=None,
        description="The size of the file in bytes"
    )
) -> str:
    """
    Format file size in bytes to human readable format.

    Args:
        size: The size of the file in bytes

    Returns:
        str: The size of the file in human readable format
    """
    if size is None or size == '0':
        return "N/A"
    size = int(size)
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if size < 1024.0:
            return f"{size:.1f} {unit}"
        size /= 1024.0
    return f"{size:.1f} PB"


def item_to_dict(
    item: dict[str, Any] = Field(
        default=None,
        description="The Google Drive item to convert to a dictionary"
    ),
    folder_prefix: str | None = Field(
        default=None,
        description="The folder name to prefix the item path with"
    )
) -> dict[str, Any]:
    """
    Convert a Google Drive item to a dictionary.

    Args:
        item: The Google Drive item to convert to a dictionary
        folder_prefix: The folder name to prefix the item path with

    Returns:
        dict[str, Any]: A dictionary containing information for the item
    """
    is_folder = item['mimeType'] == 'application/vnd.google-apps.folder'
    
    size = item.get('size', '0')
    size_str = format_size(size) if not is_folder else "N/A"
    
    owner = item.get('owners', [{}])[0].get('emailAddress', 'Unknown')
    modified_time = item.get('modifiedTime', 'Unknown')
    if modified_time != 'Unknown':
        modified_time = datetime.fromisoformat(modified_time.replace('Z', '+00:00')).strftime('%Y-%m-%d %H:%M:%S')
    
    item_dict = {
        "name": item['name'],
        "type": "folder" if is_folder else "file",
        "id": item['id'],
        "path": f"{folder_prefix}/{item['name']}" if folder_prefix else item['name'],
        "size": size_str,
        "modified": modified_time,
        "owner": owner,
        "shared": item.get('shared', False),
        "multiple_owners": len(item.get('owners', [])) > 1,
        "drive_id": item.get('driveId')
    }
    
    return item_dict



def list_directory(
    folder: str | None = Field(
        default=None,
        description="Optional ID or name of the folder to list"
    ),
    shared_drives: str | None = Field(
        default=None,
        description="Optional ID or name of shared drive to search in"
    ),
    scope: str = Field(
        default="all",
        description="Scope of files to list: 'personal' (personal files), 'shared' (files shared with me), or 'all' (both)"
    ),
    output_path: str | None = Field(
        default=None,
        description="Optional path to save the listing"
    )
) -> list[dict[str, Any]]:
    """List files and folders in Google Drive with detailed information.

    This function provides a comprehensive listing of files and folders in Google Drive,
    including both personal and shared drives. It can filter by scope and save the output
    to a file.

    Args:
        folder: ID or name of the folder to list. If None, lists root directory
        shared_drives: ID or name of shared drive to search in. If None, searches in personal drive
        scope: Scope of files to list:
            - 'personal': Only files owned by you
            - 'shared': Files shared with you or in shared drives
            - 'all': All files (default)
        output_path: Path to save the listing. If None, returns the listing as a list of dictionaries

    Returns:
        list[dict[str, Any]]: A list of dictionaries containing information for each item

    Raises:
        Exception: If there's an error accessing the drive or if no files are found
    """
    # Enable calling this step without pydantic model_validate()
    folder = folder.default if hasattr(folder, 'default') else folder
    shared_drives = shared_drives.default if hasattr(shared_drives, 'default') else shared_drives
    scope = scope.default if hasattr(scope, 'default') else scope
    output_path = output_path.default if hasattr(output_path, 'default') else output_path

    try:
        service = build("drive", "v3", credentials=load_credentials(), developerKey=load_api_key())
        output_path = Path(output_path) if output_path else None
        
        if shared_drives:
            drives = service.drives().list(pageSize=10, fields="drives(id, name)").execute()
            matching_drives = [d for d in drives.get('drives', []) if d['name'].lower() == shared_drives.lower()]
            if matching_drives:
                shared_drives = matching_drives[0]['id']
            else:
                raise Exception(f"Could not find shared drive: {shared_drives}")
        
        if folder:
            query = f"name = '{folder}' and mimeType = 'application/vnd.google-apps.folder' and trashed = false"
            params = {'q': query, 'spaces': 'drive', 'fields': "files(id, name)", 'pageSize': 1}
            if shared_drives:
                params.update({'corpora': 'drive', 'driveId': shared_drives, 'includeItemsFromAllDrives': True, 'supportsAllDrives': True})
            
            results = service.files().list(**params).execute()
            items = results.get('files', [])
            if items:
                folder = items[0]['id']
            else:
                raise Exception(f"Could not find folder: {folder}")
        
        query = "trashed = false"
        if folder:
            query += f" and '{folder}' in parents"
        
        params = {
            'q': query,
            'spaces': 'drive',
            'fields': "files(id, name, mimeType, owners, shared, size, modifiedTime, driveId)",
            'pageSize': 1000,
            'orderBy': 'name',
            'includeItemsFromAllDrives': True,
            'supportsAllDrives': True
        }
        
        if shared_drives:
            params.update({'corpora': 'drive', 'driveId': shared_drives})
        else:
            params['corpora'] = 'allDrives'
        
        results = service.files().list(**params).execute()
        items = results.get('files', [])
        
        if scope == "shared":
            items = [item for item in items if item.get('shared', False) or len(item.get('owners', [])) > 1 or item.get('driveId')]
        elif scope == "personal":
            items = [item for item in items if not item.get('shared', False) and len(item.get('owners', [])) == 1 and not item.get('driveId')]
        
        if not items:
            raise Exception("No files found.")
        
        items_list = [item_to_dict(item, folder) for item in items]
        
        if output_path:
            with open(output_path, 'w', encoding='utf-8') as f:
                json.dump(items_list, f, indent=2, ensure_ascii=False)
        else:
            return items_list
        
    except Exception as e:
        raise Exception(f"Error listing directory: {e}") from e


"""
Manages the run context for Timbal using contextvars.

This module provides a concurrency-safe way to access the current run's context.

Public API:
    - RunContext: The data model for the run context.
    - get_run_context(): Retrieves the current run context.
    - set_run_context(): Sets the current run context.

The context is typically managed automatically by the framework. Advanced users
may need to set the context manually when creating custom execution flows.
See the function docstrings for usage details and warnings.
"""
# ruff: noqa: F401
import os
from contextvars import ContextVar

from .context import RunContext, TimbalPlatformConfig

# INTERNAL: This variable holds the context. Do not access directly.
_run_context_var: ContextVar[RunContext | None] = ContextVar("run_context", default=None)


def get_run_context() -> RunContext | None:
    """
    Retrieves the current run context.

    This is the safe, recommended way to access run-specific information.

    Returns:
        The current RunContext instance, or None if no context is set.
    """
    return _run_context_var.get()


def set_run_context(context: RunContext) -> None:
    """
    Sets the run context for the current async task or thread.

    WARNING: This function is for advanced use cases, such as creating
    custom `Runnable` components or execution flows. Manually setting the
    context can lead to unexpected behavior if not handled correctly.
    """
    _run_context_var.set(context)


def resolve_platform_config() -> TimbalPlatformConfig:
    """
    Resolves the active Timbal platform configuration for the current execution context.

    This function provides a unified way to obtain the TimbalPlatformConfig used for authentication
    and resource identification in Timbal's agentic and workflow runs. It first attempts to retrieve
    the platform configuration from the current RunContext (if one is set via contextvars). If no
    RunContext is active, it falls back to constructing a configuration from environment variables:
    - TIMBAL_API_HOST (required)
    - TIMBAL_API_KEY or TIMBAL_API_TOKEN (required)

    Raises:
        ValueError: If neither a RunContext nor the required environment variables are available.

    Returns:
        TimbalPlatformConfig: The resolved platform configuration, ready for use in API calls,
        resource resolution, and authentication.

    Usage:
        Prefer this function over direct environment variable access or manual context inspection
        to ensure consistent, concurrency-safe configuration resolution throughout the Timbal framework.
    """
    current_context = get_run_context()

    if current_context and current_context.timbal_platform_config:
        return current_context.timbal_platform_config
    else:
        host = os.getenv("TIMBAL_API_HOST")
        if not host:
            raise ValueError("Missing TIMBAL_API_HOST environment variable.")
        token = os.getenv("TIMBAL_API_KEY") or os.getenv("TIMBAL_API_TOKEN")
        if not token:
            raise ValueError("Missing TIMBAL_API_KEY environment variable.")
        return TimbalPlatformConfig.model_validate({
            "host": host,
            "auth": {
                "type": "bearer",
                "token": token
            }
        })



import structlog
from pydantic import ConfigDict, field_validator

from ..validators import Validator, contains_output, not_contains_output, regex, semantic_output
from .input import Input

logger = structlog.get_logger("timbal.eval.types.output")


class Output(Input):
    """Represents the output data for a single turn in an evaluation test.

    This model defines the data associated with an agent's response or action
    within a turn. It serves a dual purpose:

    1.  **Expected Output for Validation**: When `validators` are present, this
        model's fields (e.g., `text`, `files` inherited from `Input`) define the
        expected response from the agent. The `validators` are then used to
        programmatically check if the agent's actual output matches these
        expectations.

    2.  **Fixed Record/Observation**: If the `validators` list is empty, this
        output is treated as a fixed record. This can be used to represent
        an observation or a pre-defined response to be added to the agent's
        memory or conversation history, influencing subsequent turns without
        undergoing validation itself.

    Inherits `text` and `files` attributes from the `Input` class, which define
    the content of the output.
    """
    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)

    validators: list[Validator] = []
    """A list of `Validator` instances to apply to the agent's actual output.
    If this list is empty, the `Output` instance is treated as a fixed record
    rather than an expected output to be validated.
    """


    @field_validator("validators", mode="before")
    def validate_validators(cls, v):
        """Converts validator specifications from a dictionary (e.g., YAML) into `Validator` instances."""
        if not isinstance(v, dict):
            raise ValueError("validators must be a dict")
        
        validators = []
        for validator_name, validator_arg in v.items():
            if validator_name == "contains":
                validators.append(contains_output(validator_arg))
            elif validator_name == "not_contains":
                validators.append(not_contains_output(validator_arg))
            elif validator_name == "regex":
                validators.append(regex(validator_arg))
            elif validator_name == "semantic":
                validators.append(semantic_output(validator_arg))
            # TODO Add more validators.
            else:
                logger.warning("unknown_validator", validator=validator_name)

        return validators
    
    def to_dict(self) -> dict:
        d = super().to_dict()
        
        if getattr(self, "validators", None):
            validators_dict = {}
            for v in self.validators:
                key = getattr(v, "name", str(type(v)))
                value = getattr(v, "ref", str(v))
                # Group multiple validators of the same type
                if key in validators_dict:
                    if isinstance(validators_dict[key], list):
                        validators_dict[key].append(value)
                    else:
                        validators_dict[key] = [validators_dict[key], value]
                else:
                    validators_dict[key] = value
            d["validators"] = validators_dict
        return d
    


from typing import Any

from pydantic import BaseModel

from ....types.message import Message


class LLMResult(BaseModel):
    """Helper class to wrap LLM results."""

    input: dict[str, Any]
    """The input kwargs to the LLM router."""
    output: Message | None = None
    """The output message of the LLM. Will be None if the LLM returned an error."""
    error: dict[str, Any] | None = None
    """Store if any error occurred while running the LLM."""
    t0: int 
    """The start time of the LLM in milliseconds."""
    t1: int
    """The end time of the LLM in milliseconds."""
    usage: dict[str, int] = {}
    """The usage of the LLM."""



import pytest

from timbal import Agent
from timbal.steps.perplexity import search
from timbal.core.agent.types.tool import Tool
from timbal.core.step import Step


@pytest.mark.asyncio
async def test_init_as_callable():
    agent = Agent(tools=[search])

    openai_tool_input = agent.tools[0].to_openai_tool()
    assert len(openai_tool_input["function"]["parameters"]["properties"]) == 10

    anthropic_tool_input = agent.tools[0].to_anthropic_tool()
    assert len(anthropic_tool_input["input_schema"]["properties"]) == 10


@pytest.mark.asyncio
async def test_init_as_step():
    search_step = Step(id="search", handler_fn=search)
    agent = Agent(tools=[search_step])

    openai_tool_input = agent.tools[0].to_openai_tool()
    assert len(openai_tool_input["function"]["parameters"]["properties"]) == 10

    anthropic_tool_input = agent.tools[0].to_anthropic_tool()
    assert len(anthropic_tool_input["input_schema"]["properties"]) == 10


@pytest.mark.asyncio
async def test_required_params():
    agent = Agent(
        model="gemini-2.0-flash-lite",
        tools=[{
            "runnable": search, 
            "description": "Search information on the internet.",
            "params_mode": "required",
        }],
    )

    openai_tool_input = agent.tools[0].to_openai_tool()
    assert openai_tool_input["function"]["parameters"]["properties"].keys() == {"query"}

    anthropic_tool_input = agent.tools[0].to_anthropic_tool()
    assert anthropic_tool_input["input_schema"]["properties"].keys() == {"query"}


@pytest.mark.asyncio
async def test_required_params_with_include():
    agent = Agent(
        model="gpt-4.1",
        tools=[Tool(
            runnable=search, 
            description="Search information on the internet.",
            params_mode="required",
            include_params=["model"],
        )],
    )

    openai_tool_input = agent.tools[0].to_openai_tool()
    assert openai_tool_input["function"]["parameters"]["properties"].keys() == {"query", "model"}

    anthropic_tool_input = agent.tools[0].to_anthropic_tool()
    assert anthropic_tool_input["input_schema"]["properties"].keys() == {"query", "model"}


@pytest.mark.asyncio
async def test_required_params_with_exclude():
    agent = Agent(
        model="gpt-4.1",
        tools=[Tool(
            runnable=search, 
            description="Search information on the internet.",
            exclude_params=["query"],
        )],
    )

    openai_tool_input = agent.tools[0].to_openai_tool()
    assert "query" not in openai_tool_input["function"]["parameters"]["properties"].keys()

    anthropic_tool_input = agent.tools[0].to_anthropic_tool()
    assert "query" not in anthropic_tool_input["input_schema"]["properties"].keys()


@pytest.mark.asyncio
async def test_init_invalid_dict():
    # Extra properties are ignored.
    agent = Agent(tools=[{
        "runnable": search, 
        "description": "Search information on the internet.",
        "params_mode": "required",
        "invalid_key": "invalid_value",
    }])

    # Invalid types or missing required properties raise an error.
    with pytest.raises(ValueError):
        agent = Agent(tools=[{
            "description": {},
        }])



"""
SharePoint Directory Listing Module

This module provides functionality to list and explore files and folders in SharePoint,
with support for both personal and shared files across different scopes.

Features:
---------
- List files and folders in any SharePoint location
- Support for multiple search scopes:
  - Personal files (my-files)
  - Shared files (shared)
  - All files (all)
- Detailed file/folder information:
  - Name and type
  - Size (human-readable format)
  - Creation and modification dates
  - Owner information
  - Web URL
- Optional JSON output to file
- Sorting by name
- Path-based navigation

Prerequisites:
-------------
1. Microsoft 365 Setup:
   - Create a Microsoft 365 tenant (if you don't have one)

2. App Registration:
   - Go to Azure Portal (https://portal.azure.com)
   - Navigate to "App registrations"
   - Create a new registration with:
     - Name: Your app name
     - Supported account types: Single tenant
     - Redirect URI: https://login.microsoftonline.com/common/oauth2/nativeclient
   - Configure required permissions:
     - Files.Read.All
     - Sites.Read.All

3. Authentication:
   - Run authenticate.py to get the token_sharepoint.json file
   - Ensure the token file contains valid credentials

Usage Examples:
--------------
1. List all files in personal drive:
   >>> list_directory(scope="my-files")

2. List contents of a specific folder:
   >>> list_directory(
   ...     folder="Projects/2024",
   ...     scope="my-files"
   ... )

3. List shared files:
   >>> list_directory(scope="shared")

4. Save output to JSON file:
   >>> list_directory(
   ...     scope="all",
   ...     output_path="sharepoint_contents.json"
   ... )

Note:
-----
- File/folder names are case-sensitive
- The output includes detailed metadata for each item
- Items are sorted alphabetically by name
- Empty folders are included in the listing
- Web URLs are provided for direct access to items
"""

import json
from typing import Any

from O365.drive import File, Folder

from ...types import Field
from .authenticate import get_sharepoint_account


def format_size(size: int | None) -> str:
    """Format file size in bytes to human readable format."""
    if size is None:
        return "N/A"
    
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if size < 1024.0:
            return f"{size:.1f} {unit}"
        size /= 1024.0
    return f"{size:.1f} PB"



def item_to_dict(
    item: Any = Field(
        default=None,
        description="The item to convert to a dictionary"
    ),
    folder_prefix: str | None = Field(
        default=None,
        description="The folder name to prefix the item path with"
    )
) -> dict[str, Any]:
    """Convert a SharePoint item to a dictionary."""
    if isinstance(item, Folder):
        item_type = "folder"
    elif isinstance(item, File):
        item_type = "file"
    else:
        item_type = "unknown"
    item_dict = {
        "name": item.name,
        "type": item_type,
        "path": f"{folder_prefix}/{item.name}" if folder_prefix else item.name
    }
    item_dict.update({
        "size": format_size(item.size) if item_type == "file" and hasattr(item, 'size') else "N/A",
        "modified": item.modified.strftime('%Y-%m-%d %H:%M:%S') if hasattr(item, 'modified') and item.modified else "N/A",
        "created": item.created.strftime('%Y-%m-%d %H:%M:%S') if hasattr(item, 'created') and item.created else "N/A",
        "owner": item.owner.user.display_name if hasattr(item, 'owner') and item.owner else "N/A",
        "web_url": item.web_url if hasattr(item, 'web_url') else "N/A"
    })
    return item_dict



def list_directory(
    folder: str | None = Field(
        default=None,
        description="The SharePoint folder path"
    ),
    scope: str = Field(
        default="my-files",
        description="Where to search for files: 'my-files', 'shared', or 'all'"
    ),
    output_path: str | None = Field(
        default=None,
        description="The path to save the output"
    )
) -> list[dict[str, Any]]:
    """List files and folders in SharePoint.
    
    Args:
        drive: The SharePoint drive to use
        folder: The folder path to list contents from
        scope: Where to search for files ('my-files', 'shared', or 'all')
        output_path: Path to save the output JSON file
        
    Returns:
        List of dictionaries containing file/folder information
        
    Raises:
        Exception: If there's an error accessing SharePoint or processing files
    """
    # Extract values from Field objects
    folder = folder.default if hasattr(folder, 'default') else folder
    scope = scope.default if hasattr(scope, 'default') else scope
    output_path = output_path.default if hasattr(output_path, 'default') else output_path

    account = get_sharepoint_account()
    if not account:
        raise Exception("Error: Could not get SharePoint account")
    
    # TODO: Get a specific drive
    drive = account.storage().get_default_drive()
    
    try:
        items = []

        if scope in ["my-files", "all"]:
            if folder:
                folder_item = drive.get_item_by_path(folder)
                if folder_item:
                    my_items = list(folder_item.get_items())
                    my_items.sort(key=lambda x: x.name)
                    items.extend([item_to_dict(item, folder) for item in my_items])
            else:
                my_items = list(drive.get_items())
                my_items.sort(key=lambda x: x.name)
                items.extend([item_to_dict(item, "") for item in my_items])

        if scope in ["shared", "all"]:
            shared_items = list(drive.get_shared_with_me())
            if folder:
                shared_folder = None
                for item in shared_items:
                    if isinstance(item, Folder) and item.name == folder:
                        shared_folder = item
                        break
                if shared_folder:
                    shared_inside = list(shared_folder.get_items())
                    shared_inside.sort(key=lambda x: x.name)
                    items.extend([item_to_dict(item, folder) for item in shared_inside])
            else:
                shared_root_items = [item for item in shared_items if not hasattr(item, 'parent_reference') or not getattr(item.parent_reference, 'path', None)]
                shared_root_items.sort(key=lambda x: x.name)
                items.extend([item_to_dict(item, "") for item in shared_root_items])

        if output_path:
            with open(output_path, "w", encoding='utf-8') as f:
                json.dump(items, f, indent=2, ensure_ascii=False)
        else:
            return items

    except Exception as e:
        raise Exception(f"Error listing directory: {e}") from e


from typing import Annotated

from pydantic import Field, TypeAdapter

from .copy import CopyOperation
from .delete import DeleteOperation
from .list import ListOperation
from .mkdir import MkdirOperation
from .move import MoveOperation
from .read import ReadOperation
from .write import WriteOperation

FileSystemOperation = Annotated[
    CopyOperation | DeleteOperation | ListOperation | MkdirOperation |
    MoveOperation | ReadOperation | WriteOperation,
    Field(discriminator="type"),
]
operation_adapter = TypeAdapter(FileSystemOperation)



import os

import httpx

from ...errors import APIKeyNotFoundError
from ...types.field import Field
from ...types.file import File


async def stt(
    audio_file: File = Field(
        description=(
            "The audio file to transcribe. "
            "All major audio and video formats are supported. "
            "The file size must be less than 1GB."
        ),
    ),
    model_id: str = Field(
        default="scribe_v1", 
        description="The elevenlabs model to use for the STT.",
        choices=["scribe_v1", "scribe_v1_experimental"],
    ),
    # TODO Add more fields.
) -> str: # ? Should we return the other elements.

    # Enable calling this step without pydantic model_validate()
    audio_file = audio_file.default if hasattr(audio_file, "default") else audio_file
    model_id = model_id.default if hasattr(model_id, "default") else model_id

    if not audio_file.__content_type__.startswith("audio"):
        raise ValueError(f"STT expected an audio file content type. Got {audio_file.__content_type__}")

    api_key = os.getenv("ELEVENLABS_API_KEY")
    if not api_key:
        raise APIKeyNotFoundError("ELEVENLABS_API_KEY not found")

    async with httpx.AsyncClient() as client:
        headers = {"xi-api-key": api_key}

        files = {
            "file": (
                f"audio{audio_file.__source_extension__}", 
                audio_file.read(), 
                audio_file.__content_type__
            ),
            "model_id": (None, model_id),
        }

        res = await client.post(
            "https://api.elevenlabs.io/v1/speech-to-text",
            headers=headers,
            files=files,
        )

        res.raise_for_status()

        res_body = res.json()
        # {
        #     'language_code': 'eng',
        #     'language_probability': 0.9866735935211182,
        #     'text': '...',
        #     'words': [{'text': '...', 'start': 0.079, 'end': 0.219, 'type': 'word'}, ...],
        # }
        return res_body["text"]



from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel

from ...types.field import Field, resolve_default
from ...utils import _platform_api_call


class Column(BaseModel):
    name: str
    """The name of the column."""
    data_type: str
    """The data type of the column (as a postgres data type)."""
    default_value: str | None = None
    """The default value of the column."""
    is_nullable: bool
    """Whether the column can be null."""
    is_unique: bool
    """Whether the column is unique."""
    is_primary: bool 
    """Whether the column is a primary key or part of a composite primary key."""
    comment: str | None = None
    """The comment of the column."""


class Table(BaseModel):
    name: str
    """The name of the table."""
    columns: list[Column]
    """The columns of the table."""
    comment: str | None = None
    """The comment of the table."""
    constraints: list[Any] # TODO Create a proper enum for this
    """The constraints on the table."""


async def create_table(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base to create the table in."),
    name: str = Field(description="The name of the table to create."),
    columns: list[Column] = Field(description="List of column definitions."),
    comment: str | None = Field(
        default=None,
        description="The comment of the table.",
    ),
) -> None:
    """
    Create a new table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID where the table will be created.
        name (str): The name of the table to create.
        columns (list[Column]): A list of column definitions.
        comment (str | None): An optional comment describing the table.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    name = resolve_default("name", name)
    columns = resolve_default("columns", columns)
    comment = resolve_default("comment", comment)

    # Validate columns
    columns = [column if isinstance(column, Column) else Column.model_validate(column) for column in columns]

    path = f"orgs/{org_id}/kbs/{kb_id}/tables"
    payload = {
        "name": name,
        "columns": [column.model_dump() for column in columns],
        "comment": comment,
    }

    await _platform_api_call("POST", path, json=payload)


async def delete_table(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base to delete the table from."),
    name: str = Field(description="The name of the table to delete."),
    cascade: bool = Field(description="Whether to cascade the delete to the table's indexes."),
) -> None:
    """
    Delete a table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        name (str): The name of the table to delete.
        cascade (bool): Whether to also delete all indexes and dependent objects associated with the table.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    name = resolve_default("name", name)
    cascade = resolve_default("cascade", cascade)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{name}?"
    payload = {"cascade": cascade}

    await _platform_api_call("DELETE", path, json=payload)


async def get_table(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base to get the table from."),
    name: str = Field(description="The name of the table to get."),
) -> Table:
    """
    Retrieve the full definition of a table from a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        name (str): The name of the table to retrieve.

    Returns:
        Table: A Table model containing the table's name, columns, comment, and constraints.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    name = resolve_default("name", name)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{name}"
    params = {"format": "full"}
    
    res = await _platform_api_call("GET", path, params=params)
    return Table.model_validate(res.json().get("table"))


async def get_table_sql(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base."),
    name: str = Field(description="The name of the table to get the definition for."),
) -> str:
    """
    Retrieve the complete SQL definition of a table, including its structure and constraints.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        name (str): The name of the table to retrieve the definition for.

    Returns:
        str: The full CREATE TABLE statement, including columns, constraints, and any associated indexes, as a formatted SQL string.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    name = resolve_default("name", name)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{name}"
    params = {"format": "definition"}
    
    res = await _platform_api_call("GET", path, params=params)
    return res.json().get("table")


async def get_tables(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base."),
) -> list[Table]:
    """
    List all tables in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID.

    Returns:
        list[Table]: A list of Table models, each containing the table's name, columns, comment, and constraints.
    """    
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables"
    params = {"format": "full"}
    
    res = await _platform_api_call("GET", path, params=params)
    return [Table.model_validate(table) for table in res.json().get("tables", [])]


# TODO Maybe we want to join the list into a single string (LLM oriented)
async def get_tables_sql(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base."),
) -> list[str]:
    """
    List all tables in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID.

    Returns:
        list[str]: A list of full table SQL definitions.
    """    
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables"
    params = {"format": "definition"}
    
    res = await _platform_api_call("GET", path, params=params)
    return res.json().get("tables", [])


async def import_records(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base to import records to."),
    table_name: str = Field(description="The name of the table to import records to."),
    records: list[dict[str, Any]] = Field(description="The records to import."),
) -> None:
    """
    Import records into a table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to import records to.
        records (list[dict[str, Any]]): The records to import. Each record should be a dictionary where keys match the table's column names.

    Example:
        # For an example table "Documents" with columns: id, filename, content
        await import_records(
            org_id="10",
            kb_id="48",
            table_name="Documents",
            records=[
                {"id": 1, "filename": "foo.txt", "content": "Hello world!"},
                {"id": 2, "filename": "bar.txt", "content": "Another document"}
            ]
        )
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    records = resolve_default("records", records)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}/records"
    payload = {"records": records}

    await _platform_api_call("POST", path, json=payload)


async def import_csv(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    table_name: str = Field(description="The name of the table to upload the CSV to."),
    csv_path: str | Path = Field(description="The path to the CSV file."),
    mode: Literal["append", "overwrite"] =  Field(
        default="overwrite", 
        description="The mode to use for the import.",
    ),
) -> None:
    """
    Upload a CSV file to a table in a knowledge base.

    This function imports data from a CSV file into an existing table in the specified knowledge base. 
    The CSV file must match the table's schema (column names and types). 
    You can choose to either overwrite the table's contents or append to it.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The ID of the knowledge base containing the table.
        table_name (str): The name of the table to upload the CSV to.
        csv_path (str | Path): The path to the CSV file on disk.
        mode (Literal["append", "overwrite"], optional): Import mode. Use "overwrite" to replace all existing data in the table, or "append" to add to it. Default is "overwrite".
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    csv_path = resolve_default("csv_path", csv_path)
    mode = resolve_default("mode", mode)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}/csv?mode={mode}"
    headers = {"Content-Type": "text/csv"}

    with open(csv_path, "rb") as f:
        csv_data = f.read()
    
    await _platform_api_call("POST", path, headers=headers, content=csv_data)


async def search_table(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    name: str = Field(description="The name of the table to search."),
    query: str = Field(description="The query to search for."),
    embedding_names: list[str] = Field(description="The names of the embeddings to use for the search."),
    # TODO Add more params
    limit: int = Field(
        default=10,
        description="The maximum number of results to return.",
    ),
    offset: int = Field(
        default=0,
        description="The offset to use for pagination.",
    ),
) -> list[dict[str, Any]]:
    """
    Perform a semantic search on a table within a knowledge base using embeddings.

    This function queries the specified table for records most relevant to the provided query string,
    leveraging one or more embedding models for semantic similarity. The search returns the top matching rows
    based on the embeddings and query.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The ID of the knowledge base containing the table.
        name (str): The name of the table to search.
        query (str): The natural language query or search phrase.
        embedding_names (list[str]): The names of the embedding models to use for the search.
        limit (int, optional): The maximum number of results to return. Defaults to 10.
        offset (int, optional): The offset for pagination. Defaults to 0.

    Returns:
        list[dict[str, Any]]: A list of ordered records, each as a dictionary. The structure of each record
        depends on the table schema and the specified select columns.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    name = resolve_default("name", name)
    query = resolve_default("query", query)
    embedding_names = resolve_default("embedding_names", embedding_names)
    limit = resolve_default("limit", limit)
    offset = resolve_default("offset", offset)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{name}/search"
    payload = {
        "query": query,
        "embedding_names": embedding_names,
        "limit": limit,
        "offset": offset,
    }

    res = await _platform_api_call("POST", path, json=payload)
    return res.json()


async def query(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base."),
    sql: str = Field(description="The SQL query to execute."),
) -> list[dict[str, Any]]:
    """
    Execute a SQL query against a knowledge base table (PostgreSQL dialect).

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID.
        sql (str): The SQL query to execute. This must be valid PostgreSQL SQL.

    Returns:
        list[dict[str, Any]]: The query results as a list of dictionaries, where each dictionary represents a row.

    Notes:
        - SQL syntax must follow PostgreSQL conventions.
        - Table and column names are case sensitive. If your identifiers use uppercase or mixed case, you must escape them with double quotes (e.g., "Documents", "FileName").
        - Unescaped identifiers are automatically lowercased by PostgreSQL.

    Example:
        # Count all the documents in the table
        await query(
            org_id="10",
            kb_id="48",
            sql='SELECT COUNT(*) FROM "Documents"'
        )
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    sql = resolve_default("sql", sql)

    path = f"orgs/{org_id}/kbs/{kb_id}/query"
    payload = {"sql": sql}

    res = await _platform_api_call("POST", path, json=payload)
    return res.json()
    

async def add_column(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base to create the table in."),
    table_name: str = Field(description="The name of the table to add the column to."),
    column: Column = Field(description="The column to add."),
) -> None:
    """
    Add a column to an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to add the column to.
        column (Column): The column definition to add to the table.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    column = resolve_default("column", column)

    # Validate column
    column = column if isinstance(column, Column) else Column.model_validate(column)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}"
    payload = {"operations": [{"add_column": column.model_dump()}]}

    await _platform_api_call("PATCH", path, json=payload)
    

async def drop_column(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    table_name: str = Field(description="The name of the table to drop the column from."),
    name: str = Field(description="The name of the column to drop."),
    cascade: bool = Field(description="Whether to cascade the drop to dependent objects."),
) -> None:
    """
    Drop a column from an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to drop the column from.
        name (str): The name of the column to drop.
        cascade (bool): Whether to cascade the drop to dependent objects.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    name = resolve_default("name", name)
    cascade = resolve_default("cascade", cascade)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}"
    payload = {"operations": [{"drop_column": {
        "name": name,
        "cascade": cascade
    }}]}

    await _platform_api_call("PATCH", path, json=payload)


async def rename_column(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    table_name: str = Field(description="The name of the table to rename the column in."),
    name: str = Field(description="The name of the column to rename."),
    new_name: str = Field(description="The new name for the column."),
) -> None:
    """
    Rename a column in an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to rename the column in.
        name (str): The current name of the column to rename.
        new_name (str): The new name for the column.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    name = resolve_default("name", name)
    new_name = resolve_default("new_name", new_name)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}"
    payload = {"operations": [{"rename_column": {
        "name": name,
        "new_name": new_name
    }}]}

    await _platform_api_call("PATCH", path, json=payload)


async def rename_table(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    name: str = Field(description="The name of the table to rename."),
    new_name: str = Field(description="The new name for the table."),
) -> None:
    """
    Rename an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        name (str): The current name of the table to rename.
        new_name (str): The new name for the table.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    name = resolve_default("name", name)
    new_name = resolve_default("new_name", new_name)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{name}"
    payload = {"operations": [{"rename_table": {"new_name": new_name}}]}

    await _platform_api_call("PATCH", path, json=payload)


async def add_fk(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    table_name: str = Field(description="The name of the table to add the foreign key to."),
    column_names: list[str] = Field(description="The name of the columns to add the foreign key to."),
    fk_table_name: str = Field(description="The name of the referenced table."),
    fk_column_names: list[str] = Field(description="The name of the referenced columns in the foreign table."),
    name: str = Field(description="The name of the foreign key constraint."),
    on_delete_action: str = Field(description="The action to take on delete (e.g., 'NO ACTION', 'CASCADE')."),
    on_update_action: str = Field(description="The action to take on update (e.g., 'NO ACTION', 'CASCADE')."),
) -> None:
    """
    Add a foreign key to an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to add the foreign key to.
        column_names (list[str]): The name of the columns to add the foreign key to.
        fk_table_name (str): The name of the referenced table.
        fk_column_names (list[str]): The name of the referenced columns in the foreign table.
        name (str): The name of the foreign key constraint.
        on_delete_action (str): The action to take on delete (e.g., 'NO ACTION', 'CASCADE').
        on_update_action (str): The action to take on update (e.g., 'NO ACTION', 'CASCADE').
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    column_names = resolve_default("column_names", column_names)
    fk_table_name = resolve_default("fk_table_name", fk_table_name)
    fk_column_names = resolve_default("fk_column_names", fk_column_names)
    name = resolve_default("name", name)
    on_delete_action = resolve_default("on_delete_action", on_delete_action)
    on_update_action = resolve_default("on_update_action", on_update_action)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}"
    payload = {
        "operations": [{
            "add_constraint": {
                "type": "FOREIGN KEY",
                "name": name,
                "columns": column_names,
                "reference_table": fk_table_name,
                "reference_columns": fk_column_names,
                "on_delete_action": on_delete_action,
                "on_update_action": on_update_action
            }
        }]
    }

    await _platform_api_call("PATCH", path, json=payload)


async def drop_constraint(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    table_name: str = Field(description="The name of the table to drop the constraint from."),
    name: str = Field(description="The name of the constraint to drop."),
    cascade: bool = Field(description="Whether to cascade the drop to dependent objects."),
) -> None:
    """
    Drop a constraint from an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to drop the constraint from.
        name (str): The name of the constraint to drop.
        cascade (bool): Whether to cascade the drop to dependent objects.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    name = resolve_default("name", name)
    cascade = resolve_default("cascade", cascade)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}"
    payload = {
        "operations": [{
            "drop_constraint": {
                "name": name,
                "cascade": cascade
            }
        }]
    }

    await _platform_api_call("PATCH", path, json=payload)


async def add_check(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    table_name: str = Field(description="The name of the table to add the check constraint to."),
    name: str = Field(description="The name of the check constraint."),
    expression: str = Field(description="The check constraint expression (e.g., 'length(name) < 10')."),
) -> None:
    """
    Add a check constraint to an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to add the check constraint to.
        name (str): The name of the check constraint.
        expression (str): The check constraint expression (e.g., 'length(name) < 10').
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    name = resolve_default("name", name)
    expression = resolve_default("expression", expression)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}"
    payload = {
        "operations": [{
            "add_constraint": {
                "type": "CHECK",
                "name": name,
                "expression": expression
            }
        }]
    }

    await _platform_api_call("PATCH", path, json=payload)


async def add_unique(
    org_id: str = Field(description="The organization ID."),
    kb_id: str = Field(description="The ID of the knowledge base containing the table."),
    table_name: str = Field(description="The name of the table to add the unique constraint to."),
    name: str = Field(description="The name of the unique constraint."),
    columns: list[str] = Field(description="The columns to add the unique constraint to."),
) -> None:
    """
    Add a unique constraint to an existing table in a knowledge base.

    Args:
        org_id (str): The organization ID.
        kb_id (str): The knowledge base ID containing the table.
        table_name (str): The name of the table to add the unique constraint to.
        name (str): The name of the unique constraint.
        columns (list[str]): The columns to add the unique constraint to.
    """
    org_id = resolve_default("org_id", org_id)
    kb_id = resolve_default("kb_id", kb_id)
    table_name = resolve_default("table_name", table_name)
    name = resolve_default("name", name)
    columns = resolve_default("columns", columns)

    path = f"orgs/{org_id}/kbs/{kb_id}/tables/{table_name}"
    payload = {
        "operations": [{
            "add_constraint": {
                "type": "UNIQUE",
                "name": name,
                "columns": columns
            }
        }]
    }

    await _platform_api_call("PATCH", path, json=payload)
    



from pydantic import BaseModel, ConfigDict

from .input import Input
from .output import Output
from .steps import Steps


class Turn(BaseModel):
    """Represents a single turn in an evaluation test.

    A turn defines one interaction between a user and an agent within a test case.
    Each turn specifies the input provided to the agent and the output that should result.
    The output can serve as an expected response (with validators for checking correctness)
    or as a fixed output to be used as agent memory/history (when no validators are present).
    """
    model_config = ConfigDict(extra="ignore")

    input: Input
    """The input for this turn, representing what is sent to the agent. This may include text and/or files."""
    output: Output | None = None
    """The output for this turn. When validators are present, this represents the expected output to be checked for correctness.
    When no validators are present, this output is used as a fixed response to store in the agent's memory or conversation history.
    """
    steps: Steps | None = None
    """Defines the expected sequence of tool-use or action steps for this turn, if available.
    If this field is omitted or empty, no step validation is performed for the turn.
    """

    usage: list[dict] | None = None
    """Defines the expected resource or token usage constraints for this turn, if available.
    Each constraint is represented as a dictionary, and may include `max` and/or `min` 
    values for a particular usage type (e.g., input or output tokens).
    If this field is omitted or empty, no usage constraints are applied for the turn.
    """


from pydantic import BaseModel, ConfigDict

from .turn import Turn


class Test(BaseModel):
    """Represents a single evaluation test case, identified by a unique name."""
    __test__ = False # This attribute explicitly tells pytest's discover mechanism to skip these classes.
    model_config = ConfigDict(extra="ignore")

    name: str
    """The name for the test.
    This name should be unique within the list of tests.
    """
    description: str | None = None
    """Additional description for the test."""
    turns: list[Turn]
    """The list of turns to run for this test."""



import os
import requests

from ...types.file import File
from ...errors import APIKeyNotFoundError


# TODO Async version of this.
# This right now is intended for internal use. Use at your own risk.
def download_file(private_url: str) -> File:
    slack_bot_token = os.getenv("SLACK_BOT_TOKEN")
    if not slack_bot_token:
        raise APIKeyNotFoundError("SLACK_BOT_TOKEN is not defined")

    headers = {"Authorization": f"Bearer {slack_bot_token}"}
    res = requests.get(private_url, headers=headers)

    file_bytes = res.content
    file_name = private_url.split("/")[-1]
    file_extension = "." + file_name.split(".")[-1]
    file = File.validate(file_bytes, {
        "extension": file_extension,
        "name": file_name
    })

    return file



import argparse
import asyncio
import json
import sys
from pathlib import Path

import structlog
from dotenv import find_dotenv, load_dotenv

from .. import __version__
from ..logs import setup_logging

# TODO This shouldn't be in just the server 'module'.
from ..server.utils import ModuleSpec, load_module
from .engine import eval_file
from .types.models import dump
from .types.result import EvalTestSuiteResult
from .utils import discover_files

logger = structlog.get_logger("timbal.eval")



if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Timbal evals.")
    parser.add_argument(
        "-v", 
        "--version", 
        action="store_true", 
        help="Show version and exit."
    )
    parser.add_argument(
        "--fqn",
        dest="fqn",
        type=str,
        help="Fully qualified name of the module to be run (format: path/to/file.py::object_name)",
    )
    parser.add_argument(
        "--tests",
        dest="tests",
        type=str,
        help="Path to the tests file or directory to be run.",
    )
    args = parser.parse_args()

    if args.version:
        print(f"timbal.eval {__version__}") # noqa: T201
        sys.exit(0)

    fqn = args.fqn
    if not fqn:
        print("No timbal app Fully Qualified Name (FQN) provided.") # noqa: T201
        sys.exit(1)

    fqn_parts = fqn.split("::")
    if len(fqn_parts) > 2:
        print("Invalid timbal app Fully Qualified Name (FQN) format. Use 'path/to/file.py::object_name' or 'path/to/file.py'") # noqa: T201
        sys.exit(1)
    elif len(fqn_parts) == 2:
        module_path, module_name = fqn_parts
        module_spec = ModuleSpec(
            path=Path(module_path).expanduser().resolve(), 
            object_name=module_name,
        )
    else:
        module_spec = ModuleSpec(
            path=Path(fqn_parts[0]).expanduser().resolve(), 
            object_name=None,
        )

    agent = load_module(module_spec)

    from timbal.state.savers import JSONLSaver
    agent.state_saver = JSONLSaver(path=Path("state.jsonl"))

    tests_path = args.tests
    if not tests_path:
        print("No tests path provided.") # noqa: T201
        sys.exit(1)
    if ":" in tests_path:
        tests_path, test_name = tests_path.split("::", 1)
    else:
        test_name = None
    
    tests_path = Path(tests_path).expanduser().resolve()

    if not tests_path.exists():
        print(f"Tests path {tests_path} does not exist.") # noqa: T201
        sys.exit(1)

    logger.info("loading_dotenv", path=find_dotenv())
    load_dotenv(override=True)
    setup_logging()

    files = discover_files(tests_path)
    logger.info("tests_files", files=files)

    # TODO Improve this.
    test_results = EvalTestSuiteResult()
    for file in files:
        test_results.total_tests += 1
        asyncio.run(eval_file(file, agent, test_results, test_name=test_name))

    # Save all summaries to JSON
    with open("summary.json", "w") as f:
        dumped = asyncio.run(dump(test_results))
        json.dump(dumped, f, indent=2, ensure_ascii=False)


from pathlib import Path


def discover_files(path: Path) -> list[Path]:
    """Discover all eval files given a path.
    If the path is a directory, it will search recursively for files matching the pattern "eval*.yaml".
    If the path is a file, it'll simply check the file is .yaml and return it.
    """
    files = []
    if path.is_dir():
        for file in path.rglob("eval*.yaml"):
            files.append(file)
    else:
        if not path.name.endswith(".yaml"):
            raise ValueError(f"Invalid evals path: {path}")
        files.append(path)

    return files



import os
from typing import Literal

from openai import AsyncOpenAI

from ...errors import APIKeyNotFoundError
from ...types.field import Field, resolve_default
from ...types.file import File


async def gen_images(
    prompt: str = Field(
        description="A text description of the desired image(s). Max length: 32000 characters.",
        max_length=32000,
    ),
    background: Literal["transparent", "opaque", "auto"] = Field(
        default=None,
        description="Allows to set transparency for the background of the generated image(s). "
            "Must be one of transparent, opaque or auto (default value). "
            "When auto is used, the model will automatically determine the best background for the image."
            "If transparent, the output format needs to support transparency, so it should be set to either png (default value) or webp.",
    ),
    moderation: Literal["low", "auto"] = Field(
        default=None,
        description="Control the content-moderation level for images generated by gpt-image-1. "
            "Must be either low for less restrictive filtering or auto (default value).",
    ),
    n: int = Field(
        default=None,
        description="The number of images to generate. Must be between 1 and 10.",
        ge=1,
        le=10,
    ),
    output_compression: int = Field(
        default=None,
        description="The compression level (0-100%) for the generated images. " 
            "This parameter is only supported for gpt-image-1 with the webp or jpeg output formats, and defaults to 100.",
    ),
    output_format: Literal["png", "jpeg", "webp"] = Field(
        default=None,
        description="The format in which the generated images are returned. "
            "This parameter is only supported for gpt-image-1. Must be one of png, jpeg, or webp.",
    ),
    quality: Literal["auto", "high", "medium", "low"] = Field(
        default=None,
        description="The quality of the image that will be generated.",
    ),
    size: Literal["1024x1024", "1536x1024", "1024x1536", "auto"] = Field(
        default=None,
        description="The size of the generated image.",
    ),
    # model -> we only support gpt-image-1 (the other ones are bad)
    # response_format -> we have this fixed, we return File either way
    # style -> not available for gpt-image-1
    # user -> not used
) -> list[File]:

    prompt = resolve_default("prompt", prompt)
    background = resolve_default("background", background)
    moderation = resolve_default("moderation", moderation)
    n = resolve_default("n", n)
    output_compression = resolve_default("output_compression", output_compression)
    output_format = resolve_default("output_format", output_format)
    quality = resolve_default("quality", quality)
    size = resolve_default("size", size)

    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        raise APIKeyNotFoundError("OPENAI_API_KEY not found")
    
    client = AsyncOpenAI(api_key=api_key)

    kwargs = {}

    if prompt is not None:
        kwargs["prompt"] = prompt
    if background is not None:
        kwargs["background"] = background
    if moderation is not None:
        kwargs["moderation"] = moderation
    if n is not None:
        kwargs["n"] = n
    if output_compression is not None:
        kwargs["output_compression"] = output_compression
    if output_format is not None:
        kwargs["output_format"] = output_format
    if quality is not None:
        kwargs["quality"] = quality
    if size is not None:
        kwargs["size"] = size

    res = await client.images.generate(
        model="gpt-image-1",
        **kwargs,
    )

    files = [File.validate(f"data:image/png;base64,{image.b64_json}") for image in res.data]
    return files


async def edit_images(
    images: list[File] = Field(
        description="The image(s) to edit. Must be a supported image file or an array of images. "
            "Each image should be a png, webp, or jpg file less than 50MB. You can provide up to 16 images."
    ),
    prompt: str = Field(
        description="A text description of the desired image(s). Max length: 32000 characters.",
        max_length=32000,
    ),
    background: Literal["transparent", "opaque", "auto"] = Field(
        default=None,
        description="Allows to set transparency for the background of the generated image(s). "
            "Must be one of transparent, opaque or auto (default value). "
            "When auto is used, the model will automatically determine the best background for the image."
            "If transparent, the output format needs to support transparency, so it should be set to either png (default value) or webp.",
    ),
    mask: File = Field(
        default=None,
        description="An additional image whose fully transparent areas (e.g. where alpha is zero) indicate where image should be edited. "
            "If there are multiple images provided, the mask will be applied on the first image. "
            "Must be a valid PNG file, less than 4MB, and have the same dimensions as image."
    ),
    n: int = Field(
        default=None,
        description="The number of images to generate. Must be between 1 and 10.",
        ge=1,
        le=10,
    ),
    quality: Literal["auto", "high", "medium", "low"] = Field(
        default=None,
        description="The quality of the image that will be generated.",
    ),
    size: Literal["1024x1024", "1536x1024", "1024x1536", "auto"] = Field(
        default=None,
        description="The size of the generated image.",
    ),
    # model -> we only support gpt-image-1 (the other ones are bad)
    # response_format -> we have this fixed, we return File either way
    # user -> not used
) -> list[File]:

    images = resolve_default("images", images)
    prompt = resolve_default("prompt", prompt)
    background = resolve_default("background", background)
    mask = resolve_default("mask", mask)
    n = resolve_default("n", n)
    quality = resolve_default("quality", quality)
    size = resolve_default("size", size)

    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        raise APIKeyNotFoundError("OPENAI_API_KEY not found")
    
    client = AsyncOpenAI(api_key=api_key)

    kwargs = {}

    if images is not None:
        kwargs["image"] = images
    if prompt is not None:
        kwargs["prompt"] = prompt
    if background is not None:
        kwargs["background"] = background
    if mask is not None:
        kwargs["mask"] = mask
    if n is not None:
        kwargs["n"] = n
    if quality is not None:
        kwargs["quality"] = quality
    if size is not None:
        kwargs["size"] = size

    res = await client.images.edit(
        model="gpt-image-1",
        **kwargs,
    )

    files = [File.validate(f"data:image/png;base64,{image.b64_json}") for image in res.data]
    return files



import os
import pytest
from pathlib import Path

from timbal.steps.openai.tts import tts
from timbal.types import File
from timbal.errors import APIKeyNotFoundError


# Integration tests that require real API keys
# These tests are marked separately so they can be skipped if no API key is available

@pytest.mark.asyncio
async def test_real_tts_basic():
    """Integration test for basic TTS functionality with real API."""
    
    # Skip if no API key is available
    if not os.getenv("OPENAI_API_KEY"):
        pytest.skip("OPENAI_API_KEY not available for integration test")
    
    result = await tts(text="Hello, this is a test of OpenAI text-to-speech.")
    
    # Verify the result
    assert isinstance(result, File)
    assert result.__content_type__.startswith("audio/")
    assert result.__source_extension__ == ".mp3"
    
    # Read content only once and check
    content = result.read()
    assert len(content) > 0  # Should have audio content
    assert len(content) > 100  # Audio files should be reasonably sized


@pytest.mark.asyncio
async def test_real_tts_different_voice():
    """Integration test with different voice."""
    
    if not os.getenv("OPENAI_API_KEY"):
        pytest.skip("OPENAI_API_KEY not available for integration test")
    
    result = await tts(
        text="This is a test with the Nova voice.",
        voice="nova"
    )
    
    assert isinstance(result, File)
    assert result.__content_type__.startswith("audio/")
    content = result.read()
    assert len(content) > 100


@pytest.mark.asyncio
async def test_real_tts_wav_format():
    """Integration test with WAV format."""
    
    if not os.getenv("OPENAI_API_KEY"):
        pytest.skip("OPENAI_API_KEY not available for integration test")
    
    result = await tts(
        text="Testing WAV format output.",
        response_format="wav"
    )
    
    assert isinstance(result, File)
    assert result.__source_extension__ == ".wav"
    content = result.read()
    assert len(content) > 100
    # WAV files start with "RIFF" header
    assert content[:4] == b"RIFF"


@pytest.mark.asyncio
async def test_real_tts_gpt4o_mini_with_instructions():
    """Integration test with gpt-4o-mini-tts model and instructions."""
    
    if not os.getenv("OPENAI_API_KEY"):
        pytest.skip("OPENAI_API_KEY not available for integration test")
    
    result = await tts(
        text="This message should sound very cheerful and enthusiastic!",
        model_id="gpt-4o-mini-tts",
        instructions="Speak in a very cheerful, enthusiastic, and upbeat tone."
    )
    
    assert isinstance(result, File)
    assert result.__content_type__.startswith("audio/")
    content = result.read()
    assert len(content) > 100


@pytest.mark.asyncio
async def test_real_tts_long_text():
    """Integration test with longer text content."""
    
    if not os.getenv("OPENAI_API_KEY"):
        pytest.skip("OPENAI_API_KEY not available for integration test")
    
    long_text = """
    This is a longer text to test the text-to-speech functionality with more content.
    The OpenAI TTS API should be able to handle longer passages of text without issues.
    This test helps ensure that the implementation works correctly with realistic use cases
    where users might want to convert entire paragraphs or articles to speech.
    """
    
    result = await tts(text=long_text.strip())
    
    assert isinstance(result, File)
    assert result.__content_type__.startswith("audio/")
    content = result.read()
    # Longer text should produce longer audio
    assert len(content) > 1000


@pytest.mark.asyncio 
async def test_real_tts_multiple_formats():
    """Integration test to verify different output formats work."""
    
    if not os.getenv("OPENAI_API_KEY"):
        pytest.skip("OPENAI_API_KEY not available for integration test")
    
    text = "Testing different audio formats."
    formats_to_test = ["mp3", "wav", "opus"]
    
    results = {}
    
    for format_name in formats_to_test:
        result = await tts(
            text=text,
            response_format=format_name
        )
        
        assert isinstance(result, File)
        results[format_name] = result
        
        # Verify correct extension
        expected_extension = f".{format_name}"
        assert result.__source_extension__ == expected_extension
        
        # Verify has content
        content = result.read()
        assert len(content) > 100
    
    # Different formats should produce different file sizes/content
    # (though this isn't guaranteed, it's likely)
    mp3_content = results["mp3"].read()
    wav_content = results["wav"].read()
    
    # WAV is typically uncompressed so should be larger
    assert len(wav_content) >= len(mp3_content)



import pytest

from timbal import Agent
from timbal.steps.perplexity import search
from timbal.types.events import OutputEvent, StartEvent


@pytest.mark.asyncio
async def test_nested_paths():
    agent_0 = Agent(id="agent_0", tools=[search])
    agent_1 = Agent(id="agent_1", tools=[agent_0])
    agent_2 = Agent(id="agent_2", tools=[agent_1])

    assert len(agent_2.tools) == 1
    agent_2_tool = agent_2.tools[0]
    assert agent_2_tool.path == "agent_2.agent_1"

    assert len(agent_2_tool.tools) == 1
    agent_1_tool = agent_1.tools[0]
    assert agent_1_tool.path == "agent_2.agent_1.agent_0"

    assert len(agent_1_tool.tools) == 1
    agent_0_tool = agent_0.tools[0]
    assert agent_0_tool.path == "agent_2.agent_1.agent_0.search"
        

@pytest.mark.asyncio
async def test_nested_1():
    weather_expert = Agent(
        id="weather_expert",
        tools=[{
            "runnable": search,
            "description": "Search the internet for any information.",
            "params_mode": "required",
        }],
    )

    agent = Agent(tools=[weather_expert])

    events = []
    async for event in agent.run(prompt="What is the weather like in New York?"):
        # Hack for being able to check the events in the test easily.
        if isinstance(event, StartEvent):
            events.append(f"START:{event.path}")
        elif isinstance(event, OutputEvent):
            events.append(f"OUTPUT:{event.path}")

    assert "START:agent" in events
    assert "START:agent.llm-0" in events
    assert "OUTPUT:agent.llm-0" in events
    assert any(event.startswith("START:agent.weather_expert-") for event in events)
    assert any(event.startswith("OUTPUT:agent.weather_expert-") for event in events)
    assert "START:agent.llm-1" in events
    assert "OUTPUT:agent.llm-1" in events
    assert "OUTPUT:agent" in events

    # Hence nested events are not yielded by the parent agent.
    # TODO Think what to do if we want to yield them.



import json
import os
from typing import Any

import httpx

from ...types.field import Field


async def send_whatsapp_message(
    to: str = Field(description="The WhatsApp account to send the message to in E.164 format"),
    message: str = Field(
        description=(
            "The message to send. "
            "The maximum length is 1600 characters. "
            "If the message is longer, it will be split into multiple messages."
        ),
    ),
) -> None:
    """A WhatsApp session begins when a user sends a message to your app. 
    Sessions are valid for 24 hours after the most recently received message, 
    during which you can communicate with customers using free-form messages. 
    To send a message outside the 24-hour session window, you must use a pre-approved message template.
    """

    twilio_account_sid = os.getenv("TWILIO_ACCOUNT_SID")
    twilio_auth_token = os.getenv("TWILIO_AUTH_TOKEN")
    twilio_from = os.getenv("TWILIO_FROM")
    
    url = f"https://api.twilio.com/2010-04-01/Accounts/{twilio_account_sid}/Messages.json"

    data = {
        "To": f"whatsapp:{to}",
        "From": f"whatsapp:{twilio_from}",
        "Body": message
    }
    
    async with httpx.AsyncClient() as client:
        response = await client.post(
            url,
            data=data,
            auth=(twilio_account_sid, twilio_auth_token)
        )
        response.raise_for_status()


async def send_whatsapp_template(
    to: str = Field(description="The WhatsApp account to send the message to in E.164 format"),
    template_sid: str = Field(description="The template SID to send the message to"),
    template_params: dict[str, Any] = Field(
        description=(
            "The parameters to send to the template. "
            "The parameters must match the template parameters exactly."
        ),
    ),
) -> None:
    """A WhatsApp session begins when a user sends a message to your app. 
    Sessions are valid for 24 hours after the most recently received message, 
    during which you can communicate with customers using free-form messages. 
    To send a message outside the 24-hour session window, you must use a pre-approved message template.
    """

    twilio_account_sid = os.getenv("TWILIO_ACCOUNT_SID")
    twilio_auth_token = os.getenv("TWILIO_AUTH_TOKEN")
    twilio_from = os.getenv("TWILIO_FROM")
    
    url = f"https://api.twilio.com/2010-04-01/Accounts/{twilio_account_sid}/Messages.json"

    data = {
        "To": f"whatsapp:{to}",
        "From": f"whatsapp:{twilio_from}",
        "ContentSid": template_sid,
        "ContentVariables": json.dumps(template_params),
    }
    
    async with httpx.AsyncClient() as client: # TODO Think timeout
        response = await client.post(
            url,
            data=data,
            auth=(twilio_account_sid, twilio_auth_token)
        )
        response.raise_for_status()



import os

import httpx

from ...errors import APIKeyNotFoundError
from ...types.field import Field
from ...types.file import File


async def tts(
    text: str = Field(
        description="The text to convert to speech.",
    ),
    voice_id: str = Field(
        description="The voice ID to use for text-to-speech.",
    ),
    model_id: str = Field(
        default="eleven_flash_v2_5",
        description="The model to use for text-to-speech.",
        choices=["eleven_flash_v2_5", "eleven_multilingual_v2"],
    ),
    # TODO Add more fields.
) -> File:

    # Enable calling this step without pydantic model_validate()
    text = text.default if hasattr(text, "default") else text
    voice_id = voice_id.default if hasattr(voice_id, "default") else voice_id
    model_id = model_id.default if hasattr(model_id, "default") else model_id

    api_key = os.getenv("ELEVENLABS_API_KEY")
    if not api_key:
        raise APIKeyNotFoundError("ELEVENLABS_API_KEY not found")

    async with httpx.AsyncClient() as client:
        headers = {"xi-api-key": api_key}

        data = {
            "text": text,
            "model_id": model_id,
        }

        res = await client.post(
            f"https://api.elevenlabs.io/v1/text-to-speech/{voice_id}?output_format=mp3_44100_128",
            headers=headers,
            json=data,
        )

        res.raise_for_status()

        return File.validate(
            res.content, 
            {"extension": ".mp3"}
        )



import structlog
from pydantic import BaseModel, ConfigDict, field_validator

from ..validators import Validator, contains_steps, not_contains_steps, regex, semantic_steps

logger = structlog.get_logger("timbal.eval.types.steps")


class Steps(BaseModel):
    """Represents the steps data for a single turn in an evaluation test.

    This model defines the expected sequence of tool-use or action steps that an agent should perform within a turn.
    It serves a dual purpose:

    1.  **Expected Steps for Validation**: When `validators` are present, this model's fields define the
        expected steps (such as tool invocations or actions) that the agent should execute. The `validators`
        are then used to programmatically check if the agent's actual steps match these expectations.

    2.  **Fixed Record/Observation**: If the `validators` list is empty, this steps instance is treated as a
        fixed record, representing a sequence of actions or steps to be added to the agent's memory or
        conversation history, influencing subsequent turns without undergoing validation itself.

    The `validators` attribute specifies how the actual steps are to be checked against the expected steps.
    """
    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)

    validators: list[Validator] = []
    """A list of `Validator` instances to apply to the agent's actual steps.
    If this list is empty, the `Steps` instance is treated as a fixed record
    rather than an expected sequence of steps to be validated.
    """


    @field_validator("validators", mode="before")
    def validate_validators(cls, v):
        """Converts validator specifications from a dictionary (e.g., YAML) into `Validator` instances."""
        if not isinstance(v, dict):
            raise ValueError("validators must be a dict")
        
        validators = []
        for validator_name, validator_arg in v.items():
            if validator_name == "contains":
                validators.append(contains_steps(validator_arg))
            # elif validator_name == "regex":
            #     validators.append(regex(validator_arg))
            elif validator_name == "semantic":
                validators.append(semantic_steps(validator_arg))
            elif validator_name == "not_contains":
                validators.append(not_contains_steps(validator_arg))
            # TODO Add more validators.
            else:
                logger.warning("unknown_validator", validator=validator_name)

        return validators
        
    def to_dict(self) -> dict:
        d = {}
        
        if getattr(self, "validators", None):
            validators_dict = {}
            for v in self.validators:
                key = getattr(v, "name", str(type(v)))
                value = getattr(v, "ref", str(v))
                # Group multiple validators of the same type
                if key in validators_dict:
                    if isinstance(validators_dict[key], list):
                        validators_dict[key].append(value)
                    else:
                        validators_dict[key] = [validators_dict[key], value]
                else:
                    validators_dict[key] = value
            d["validators"] = validators_dict
        return d
    